from backtrader import BrokerBase, OrderBase, Order
from backtrader.position import Position
from backtrader.utils.py3 import queue, with_metaclass
import random
from xtquant.xttrader import XtQuantTrader
from xtquant.xttype import StockAccount
from .qmtstore import QMTStore

class QMTOrder(OrderBase):
    def __init__(self, owner, data, ccxt_order):
        self.owner = owner
        self.data = data
        self.ccxt_order = ccxt_order
        self.executed_fills = []
        self.ordtype = self.Buy if ccxt_order['side'] == 'buy' else self.Sell
        self.size = float(ccxt_order['amount'])

        super(QMTOrder, self).__init__()

class MetaQMTBroker(BrokerBase.__class__):
    def __init__(cls, name, bases, dct):
        '''Class has already been created ... register'''
        # Initialize the class
        super(MetaQMTBroker, cls).__init__(name, bases, dct)
        QMTStore.BrokerCls = cls

class QMTBroker(BrokerBase, metaclass=MetaQMTBroker):
    pass

    def __init__(self, **kwargs):

        self.store = QMTStore(**kwargs)
        self.mini_qmt_path = kwargs.get('mini_qmt_path')
        self.account_id = kwargs.get('account_id')

        session_id = int(random.randint(100000, 999999))

        xt_trader = XtQuantTrader(self.mini_qmt_path, session_id)
        # 启动交易对象
        xt_trader.start()
        # 连接客户端
        connect_result = xt_trader.connect()

        if connect_result == 0:
            print('连接成功')

        account = StockAccount(self.account_id)
        # 订阅账号
        res = xt_trader.subscribe(account)

        self.xt_trader = xt_trader
        self.account = account


    def getcash(self):
        res = self.query_stock_asset(self.account)

        self.cash = res.cash

        return self.cash
    
    def getvalue(self, datas=None):

        res = self.query_stock_asset(self.account)

        self.value = res.market_value

        return self.value
    
    def getposition(self, data, clone=True):

        xt_position = self.xt_trader.query_stock_position(self.account, data._dataname)
        pos = Position(size=xt_position.volume, price=xt_position.avg_price)
        return pos
    
    def get_notification(self):
        try:
            return self.notifs.get(False)
        except queue.Empty:
            return None
    
    def notify(self, order):
        self.notifs.put(order)

    def next(self):
        pass

    def buy(self, owner, data, size, price=None, plimit=None,
            exectype=None, valid=None, tradeid=0, oco=None,
            trailamount=None, trailpercent=None,
            **kwargs):
        pass

    def sell(self, owner, data, size, price=None, plimit=None,
             exectype=None, valid=None, tradeid=0, oco=None,
             trailamount=None, trailperc7ent=None,
             **kwargs):
        pass

    def cancel(self, order):
        pass




import random
from xtquant import xtdata, xttrader, xttype
from backtrader.metabase import MetaParams
import backtrader as bt
import pandas as pd


class MetaSingleton(MetaParams):
    '''Metaclass to make a metaclassed class a singleton'''

    def __init__(cls, name, bases, dct):
        super(MetaSingleton, cls).__init__(name, bases, dct)
        cls._singleton = None

    def __call__(cls, *args, **kwargs):
        if cls._singleton is None:
            cls._singleton = (
                super(MetaSingleton, cls).__call__(*args, **kwargs))

        return cls._singleton


class QMTStore(object, metaclass=MetaSingleton):
    
    def getdata(self, *args, **kwargs): 
        '''Returns ``DataCls`` with args, kwargs'''
        kwargs['store'] = self
        qmtFeed = self.__class__.DataCls(*args, **kwargs)
        return qmtFeed
    
    def getdatas(self, *args, **kwargs):
        '''Returns ``DataCls`` with *args, **kwargs (multiple entries)'''
        return [self.getdata(*args, **{**kwargs, 'dataname': stock}) for stock in kwargs.pop('code_list', 1)]
    
    def setdatas(self, cerebro, datas):
        '''Set the datas'''
        for data in datas:
            cerebro.adddata(data)

    def getbroker(self, *args, **kwargs):
        '''Returns broker with *args, **kwargs from registered ``BrokerCls``'''
        return self.__class__.BrokerCls(*args, **kwargs)
    
    def __init__(self):

        self.mini_qmt_path = ''
        self.code_list = []
        self.last_tick = None
        self.token = None

    def _get_benchmark(self):
        xtdata.download_history_data(stock_code='000300.SH', period='1d', start_time='2022-01-01', end_time='2023-01-01', dividend_type='none')
        pass
    
    def connect(self, mini_qmt_path, account):

        try:
            xtdata.connect()
        except:
            return -1

        session_id = int(random.randint(100000, 999999))
        xt_trader = xttrader.XtQuantTrader(mini_qmt_path, session_id)

        xt_trader.start()

        connect_result = xt_trader.connect()

        if connect_result == 0:
            print('连接成功')

            self.stock_account = xttype.StockAccount(account)

            xt_trader.subscribe(self.stock_account)

            self.xt_trader = xt_trader
        
        return connect_result

    def _auto_expand_array_columns(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Write by ChatGPT4

        Automatically identify and expand DataFrame columns containing array values.

        Returns:
        - A new DataFrame with the expanded columns.
        """
        for col in df.columns:
            if df[col].apply(lambda x: isinstance(x, (list, tuple))).all():
                # Expand the array column into a new DataFrame
                expanded_df = df[col].apply(pd.Series)
                # Generate new column names
                expanded_df.columns = [f"{col}{i+1}" for i in range(expanded_df.shape[1])]
                # Drop the original column and join the expanded columns
                df = df.drop(col, axis=1).join(expanded_df)
        return df

    def _fetch_history(self, symbol, period, start_time='', end_time='', count=-1, dividend_type='none', download=True):
        """
        获取历史数据
        
        参数：
            symbol: 标的代码
            period: 周期
            start_time: 起始日期
            end_time: 终止日期

        """
        if download:
            xtdata.download_history_data(stock_code=symbol, period=period, start_time=start_time, end_time=end_time)
        res = xtdata.get_market_data_ex(stock_list=[symbol], period=period, start_time=start_time, end_time=end_time, count=count, dividend_type=dividend_type)
        res = res[symbol]
        if period == 'tick':
            res = self._auto_expand_array_columns(res)
        return res
    
    def _subscribe_live(self, symbol, period, callback, start_time='', end_time=''):

        seq = xtdata.subscribe_quote(stock_code=symbol, period=period, start_time=start_time, end_time=end_time, callback=callback)

        return seq

    
    def _unsubscribe_live(self, seq):
        xtdata.unsubscribe_quote(seq)



#!/usr/bin/env python
from __future__ import (absolute_import, division, print_function,
                        unicode_literals)

from collections import deque
import datetime
import backtrader as bt
from backtrader.feed import DataBase
import time
import threading
import random

from .qmtstore import QMTStore

class MetaQMTFeed(DataBase.__class__):
    def __init__(cls, name, bases, dct):
        '''Class has already been created ... register'''
        # Initialize the class
        super(MetaQMTFeed, cls).__init__(name, bases, dct)

        # Register with the store
        QMTStore.DataCls = cls


class QMTFeed(DataBase, metaclass=MetaQMTFeed):
    """
    QMT eXchange Trading Library Data Feed.
    Params:
      - ``historical`` (default: ``False``)
    """

    lines = ('lastClose', 'amount', 'pvolume', 'stockStatus', 'openInt', 'lastSettlementPrice', 'settlementPrice', 'transactionNum', 'askPrice1', 'askPrice2', 'askPrice3', 'askPrice4', 'askPrice5', 'bidPrice1', 'bidPrice2', 'bidPrice3', 'bidPrice4', 'bidPrice5', 'askVol1', 'askVol2', 'askVol3', 'askVol4', 'askVol5', 'bidVol1', 'bidVol2', 'bidVol3', 'bidVol4', 'bidVol5', 'openInterest', 'dr', 'totaldr', 'preClose', 'suspendFlag', 'settelementPrice', 'pe' )

    params = (
        ('live', False),  # only historical download
        ('timeframe', bt.TimeFrame.Ticks)
    )

    def __init__(self, **kwargs):
        self._timeframe = self.p.timeframe
        self._compression = 1
        self.store = kwargs['store']
        # self.cerebro = kwargs['cerebro']
        self._data = deque()  # data queue for price data
        self._seq = None

    # def __len__(self):
    #     return len(self._data)

    def start(self, ):
        DataBase.start(self)

        period_map = {
            bt.TimeFrame.Days: '1d',
            bt.TimeFrame.Minutes: '1m',
            bt.TimeFrame.Ticks: 'tick'
        } 

        if not self.p.live:
            self._history_data(period=period_map[self.p.timeframe])
            print(f'{self.p.dataname}历史数据装载成功！')
        else:
            self._live_data(period=period_map[self.p.timeframe])
            print(f'{self.p.dataname}实时数据装载成功！')

    def stop(self):
        DataBase.stop(self)

        if self.p.live:
            self.store._unsubscribe_live(self._seq)

    def _get_datetime(self, value):
        dtime = datetime.datetime.fromtimestamp(value // 1000)
        return bt.date2num(dtime)

    def _load_current(self, current):
        for key in current.keys():   
                try: 
                    value = current[key]
                    if key == 'time':
                        self.lines.datetime[0] = self._get_datetime(value)
                    elif key == 'lastPrice' and self.p.timeframe == bt.TimeFrame.Ticks:
                        self.lines.close[0] = value
                    else:
                        attr = getattr(self.lines, key)
                        attr[0] = value
                except Exception as e:
                    print(e)
                    pass
        # print(current, 'current')
        self.put_notification(int(random.randint(100000, 999999)))

    def _load(self, replace=False):
        if len(self._data) > 0:

            current = self._data.popleft()

            self._load_current(current)

            return True
        return None
    
    def haslivedata(self):
        return self.p.live and self._data

    def islive(self):
        return self.p.live
    
    def _format_datetime(self, dt, period='1d'):
        if dt is None:
            return ''
        else:
            if period == '1d':
                formatted_string = dt.strftime("%Y%m%d")
            else:
                formatted_string = dt.strftime("%Y%m%d%H%M%S")
            return formatted_string

    def _append_data(self, item):
        self._data.append(item)    
    
    def _history_data(self, period):

        start_time = self._format_datetime(self.p.fromdate, period)
        end_time = self._format_datetime(self.p.todate, period)

        res = self.store._fetch_history(symbol=self.p.dataname, period=period, start_time=start_time, end_time=end_time)
        result = res.to_dict('records')
        for item in result:
            if item.get('close') != 0 and item.get('lastPrice') != 0:
                self._data.append(item)

    def _live_data(self, period):

        start_time = self._format_datetime(self.p.fromdate, period)

        def on_data(res):
            print(self.lines.datetime)
            # current = res[self.p.dataname][0]
            # if self._get_datetime(current['time']) == self.lines.datetime[0]:
            #     self._load_current(current)
            # else:
            #     self._data.append(current)


        self._seq = self.store._subscribe_live(symbol=self.p.dataname, period=period, start_time=start_time, callback=on_data)

        res = self.store._fetch_history(symbol=self.p.dataname, period=period, start_time=start_time, download=False)
        result = res.to_dict('records')
        for item in result:
            self._data.append(item)

    # def test(self, close):
    #     # 获取当前日期
    #     current_date = datetime.datetime.now().date()

    #     # 将当前日期转换为datetime对象，并将时间设置为00:00:00
    #     start_of_day = datetime.datetime.combine(current_date, datetime.time.min)

    #     # 将开始时间转换为Unix时间戳（以毫秒为单位）
    #     start_of_day_unix_ms = int(start_of_day.timestamp() * 1000)
    #     current = {
    #         'time': start_of_day_unix_ms,
    #         'close': close,
    #         'open': 1.0,
    #         'high': 1.0,
    #         'low': 1.0,
    #         'volume': 1.0,
    #     }
    #     self._load_current(current)

    # # 定义一个函数，这个函数将在每次循环中执行
    # def my_function():
    #     while True:
    #         self.test(1)
    #         time.sleep(3)  # 暂停3秒

    # # 创建并启动线程
    # thread = threading.Thread(target=my_function)
    # thread.daemon = True  # 设置为守护线程，这样主线程结束时，子线程也会结束
    # thread.start()

